package de.unikoblenz.fgbks.core.dmn.verification.verifier.impl;

import static de.unikoblenz.fgbks.base.domain.Name.NO_NAME;
import static de.unikoblenz.fgbks.core.dmn.domain.vdmn.utils.VDmnFunctions.getDecisionStringName;
import static de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntry.VerificationClassification.WARNING;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.Action.ACTION_SHOW_DECISION;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.Action.ACTION_SHOW_INPUT_DATA;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionScope.INPUT_COLUMN;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionScope.INPUT_DATA;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionScope.OUTPUT_COLUMN;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionType.CREATE;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionType.DELETE;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.VerificationFix.DEFAULT_SHOW_NAME;

import de.unikoblenz.fgbks.base.domain.Name;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecision;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputData;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnNode;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnOutputColumn;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntryElement;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.Action;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.VerificationFix;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.AbstractVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.DmnVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.MissingInputColumnVerification;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

@DmnVerifier(verifierType = MissingInputColumnVerification.class)
public class MissingInputColumnVerifier extends AbstractVerifier {

  @Override
  protected void doVerification() {
    // 1. check for input data nodes
    for (VDmnInputData inputData : dmnObjectContainer.getVDmnDefinition().getVDmnInputData()) {
      if (inputData.getName().isPresent()) {
        List<Name> inputDataNames = Collections.singletonList(inputData.getName().get());
        checkInputData(inputData, inputDataNames);
      }
    }
    // 2. check for decision nodes
    for (VDmnDecision decision : dmnObjectContainer.getVDmnDefinition().getVDmnDecisions()) {
      List<Name> inputDataNames =
          decision.getVDmnDecisionTable().getVDmnOutputColumns().stream()
              .map(c -> c.getName().orElse(null))
              .filter(Objects::nonNull)
              .collect(Collectors.toList());
      checkInputData(decision, inputDataNames);
    }
  }

  private void checkInputData(VDmnNode inputDataNode, List<Name> inputDataNames) {
    // Get all outgoing decisions from the current input node
    outer:
    for (VDmnDecision decision : inputDataNode.getInformationProvidingDecisions()) {
      // Check all input columns of the decision node
      List<Name> inputColumnNames =
          decision.getVDmnDecisionTable().getVDmnInputColumns().stream()
              .map(x -> x.getName().orElse(null))
              .filter(Objects::nonNull)
              .collect(Collectors.toList());
      for (Name inDataName : inputDataNames) {
        if (inputColumnNames.contains(inDataName)) {
          continue outer;
        }
      }
      vreFactory
          .addElement(VerificationResultEntryElement.create(inputDataNode))
          .addElement(VerificationResultEntryElement.create(decision))
          .addVerificationFix(
              VerificationFix.getBuilder()
                  .withFixName(DEFAULT_SHOW_NAME)
                  .addAction(ACTION_SHOW_INPUT_DATA)
                  .addAction(ACTION_SHOW_DECISION)
                  .build());
      // Add action remove input data
      if (inputDataNode.isInputData()) {
        vreFactory.addVerificationFix(
            VerificationFix.getBuilder()
                .withFixName(new Name("Delete Input Data"))
                .addAction(
                    Action.getBuilder()
                        .addIdValue(inputDataNode.getId())
                        .withActionType(DELETE)
                        .withActionScope(INPUT_DATA)
                        .build())
                .build());
      }
      // Add action add input column
      if (inputDataNode.isInputData()) {
        vreFactory.addVerificationFix(
            VerificationFix.getBuilder()
                .withFixName(
                    new Name("Add Input Column " + inputDataNode.getName().get().getValue()))
                .addAction(
                    Action.getBuilder()
                        .withActionType(CREATE)
                        .withActionScope(INPUT_COLUMN)
                        .addIdValue(decision.getDecisionId())
                        .addValue("name", inputDataNode.getName().get().getValue())
                        .addValue("typeRef", "string")
                        .build())
                .build());
      } else {
        for (VDmnOutputColumn outputColumn :
            ((VDmnDecision) inputDataNode).getVDmnDecisionTable().getVDmnOutputColumns()) {
          if (outputColumn.getName().isPresent()) {
            // add input Column
            vreFactory.addVerificationFix(
                VerificationFix.getBuilder()
                    .withFixName(
                        new Name("Add Input Column \"" + outputColumn.getName().get().getValue())
                            + "\"")
                    .addAction(
                        Action.getBuilder()
                            .withActionType(CREATE)
                            .withActionScope(INPUT_COLUMN)
                            .addIdValue(decision.getDecisionId())
                            .addValue("name", outputColumn.getName().get().getValue())
                            .addValue("typeRef", outputColumn.getTypeRef().getName())
                            .build())
                    .build());
          }
          vreFactory.addVerificationFix(
              VerificationFix.getBuilder()
                  .withFixName(
                      new Name(
                          "Delete Output Column \""
                              + outputColumn.getName().orElse(NO_NAME).getValue())
                          + "\"")
                  .addAction(
                      Action.getBuilder()
                          .withActionType(DELETE)
                          .withActionScope(OUTPUT_COLUMN)
                          .addIdValue(outputColumn.getVDmnDecision().getDecisionId())
                          .addIdValue(outputColumn.getOutputId())
                          .build())
                  .build());
        }
      }

      vreFactory.addToEntry(
          WARNING,
          (inputDataNode.isInputData() ? "Input data" : "Decision")
              + " node \"%s\" has no corresponding input column in decision \"%s\"",
          inputDataNode.getName().orElse(NO_NAME),
          getDecisionStringName(decision));
    }
  }
}
